/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Id: klFixedBase_arm.h,v 1.5 2004/08/22 13:42:13 wasserer Exp $
 *
\**************************************************************************/

//
// optimized MATHBASE implementation for ARM processors, also
// works with emulator builds. Only 16.16 bit precission 
// is supported.
//
// many operations of this MATHBASE are adapted from the
//

/*
 * GLESonGL implementation
 * Version:  1.0
 *
 * Copyright (C) 2003  David Blythe   All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 */

#ifndef __KLFIXEDBASE_ARM_HEADERFILE__
#define __KLFIXEDBASE_ARM_HEADERFILE__


class klFixedBase_arm
{
public:
	enum {
		PBITS = 16
	};

	const static int ZERO;
	const static int ONE;
	const static int X_2PI;
	const static int X_R2PI;
	static const unsigned short __gl_sin_tab[];

	static float floatFromFixed(int nFixed)
	{
		return nFixed/(float)(1 << PBITS);
	}

	static double doubleFromFixed(int nFixed)
	{
		return nFixed/(double)(1 << PBITS);
	}

	static int fixedFromInt(int nV)
	{
		return nV<<PBITS;
	}

	static int fixedFromFloat(float nV)
	{
		return (int)(nV * (1 << PBITS) + 0.5f);
	}

	static int fixedFromDouble(double nV)
	{
		return (int)(nV * (1 << PBITS) + 0.5f);
	}

	static int inverse(int a)
	{
		// adopted from GLESonGL math.c, __gl_f_recip(__gl_f a)

		int exp, i;
		int x;
		/* 1/(4x) */
		static const unsigned short __gl_rcp_tab[] = { /* domain 0.5 .. 1.0-1/16 */
			0x8000, 0x71c7, 0x6666, 0x5d17, 0x5555, 0x4ec4, 0x4924, 0x4444
		};
		if (a == ZERO) return 0x7fffffff;
		x = a;
		exp = 31;
		if (x & 0xffff0000) { exp -= 16; x >>= 16; }
		if (x & 0xff00) { exp -= 8; x >>= 8; }
		if (x & 0xf0) { exp -= 4; x >>= 4; }
		if (x & 0xc) { exp -= 2; x >>= 2; }
		if (x & 0x2) { exp -= 1; }
		x = __gl_rcp_tab[(a>>(28-exp))&0x7]<<2;
		exp -= 16;
		if (exp <= 0)
		x >>= -exp;
		else
		x <<= exp;

		/* two iterations of newton-raphson  x = x(2-ax) */
		i = 2;
		do {
			x = multiply(x,(2*ONE - multiply(a,x)));
		} while (--i > 0);

		return x;
	}

	static int multiply(int nLeftFixed, int nRightFixed)
	{
		#ifndef __MARM_ARMI__
			// Symbian does not provide a primitive 64 bit data type 
			// like MS __int64. Instead the class TInt64 should
			// be used. To avoid dependencies with complex Symbian 
			// includes and libraries the 64 bit multiplication is 
			// implemented from scratch. This code is used in 
			// emulator builds only.
			int u = nLeftFixed;
			int v = nRightFixed;

			int      u0, v0;
			unsigned u1, v1, t;
			unsigned w1, w2;

			// "emulate" 32x32 -> 64 bit multiplication
			u0 = u >> 16; u1 = u & 0xFFFF;
			v0 = v >> 16; v1 = v & 0xFFFF;

			t = u0*v1 + ((u1*v1) >> 16);
			w2 = t & 0xFFFF;
			w1 = static_cast<int>(t) >> 16;

			int hi = u0*v0 + w1 + ((u1*v0 + w2) >> 16);
			unsigned lo = u*v;

			return ( hi<<16 ) | ( lo>>16 );
		#else
			// Do 64 bit multiplication in a single
			// ARM command
			int lo,hi;
			asm volatile("smull\t%0, %1, %2, %3"
						   : "=&r"(lo),"=&r"(hi)
						   : "%r"(nLeftFixed),"r"(nRightFixed)
						   : "cc");
			return ( hi<<16 ) | ( 0xffff & ( lo>>16 ) );
		#endif
	}

	static int divide(int nLeftFixed, int nRightFixed)
	{
		// not as accurate as a real division, but much faster
		return multiply(nLeftFixed, inverse(nRightFixed));
	}

	static int cos(int a)
	{
		int v;

		/* reduce to [0,1) */
		while (a < ZERO) a += X_2PI;
		a *= X_R2PI;
		a >>= PBITS;
		a += 0x4000;

		/* now in the range [0, 0xffff], reduce to [0, 0xfff] */
		a >>= 4;

		v = (a & 0x400) ? __gl_sin_tab[0x3ff - (a & 0x3ff)] : __gl_sin_tab[a & 0x3ff];
		v = multiply(v,ONE);
		return (a & 0x800) ? -v : v;
	}

	static int sin(int a)
	{
		int v;

		/* reduce to [0,1) */
		while (a < ZERO) a += X_2PI;
		a *= X_R2PI;
		a >>= PBITS;

		/* now in the range [0, 0xffff], reduce to [0, 0xfff] */
		a >>= 4;

		v = (a & 0x400) ? __gl_sin_tab[0x3ff - (a & 0x3ff)] : __gl_sin_tab[a & 0x3ff];
		v = multiply(v,ONE);
		return (a & 0x800) ? -v : v;
	}

	static int fabs(int nFixed)
	{
		return nFixed<0 ? -nFixed : nFixed;
	}

	static int sqrt(int a)
	{
		int x /*= (a+__GL_X_ONE)>>1*//*a>>1*/;
		/* 1/(2sqrt(x)) - extra divide by 2 to scale to 16 bits */
		static const unsigned short __gl_rsq_tab[] = { /* domain 0.5 .. 1.0-1/16 */
		0xb504, 0xaaaa, 0xa1e8, 0x9a5f, 0x93cd, 0x8e00, 0x88d6, 0x8432,
		};
		int i, exp;
		if (a == ZERO) return 0x7fffffff;
		if (a == ONE) return a;
		x = a;
		exp = 31;
		if (x & 0xffff0000) { exp -= 16; x >>= 16; }
		if (x & 0xff00) { exp -= 8; x >>= 8; }
		if (x & 0xf0) { exp -= 4; x >>= 4; }
		if (x & 0xc) { exp -= 2; x >>= 2; }
		if (x & 0x2) { exp -= 1; }
		x = __gl_rsq_tab[(a>>(28-exp))&0x7]<<1;

		exp -= 16;
		if (exp <= 0)
			x >>= -exp>>1;
		else
			x <<= (exp>>1)+(exp&1);
		if (exp&1) x = multiply(x, __gl_rsq_tab[0]);

		/* newton-raphson */
		/* x = x/2*(3-(a*x)*x) */
		i = 0;
		do {
			x = multiply((x>>1),(3*ONE - multiply(multiply(a,x),x)));
		} while(++i < 3);

		return x;
	}

	static int inverseSqrt(int nFixed)
	{
		return inverse(sqrt(nFixed));
	}

	static int ceil(int nFixed)
	{
		int ret = (nFixed>>PBITS)<<PBITS;

		if(nFixed>=0 && ret<nFixed)
			ret += fixedFromInt(1);

		return ret;
	}

};

#ifdef _KLMATH_CPP_
	// define static members only if included from klMath.cpp

	const int klFixedBase_arm::ZERO = 0;
	const int klFixedBase_arm::ONE  = 1 << PBITS;
	const int klFixedBase_arm::X_2PI  = static_cast<int>((2*3.1415926535897932384626433832795f)*(1<<16));
	const int klFixedBase_arm::X_R2PI = static_cast<int>((1/(2*3.1415926535897932384626433832795f)*(1<<16)));

	const unsigned short klFixedBase_arm::__gl_sin_tab[] = {
		#include "gles/gl_sin.h"
	};
#endif

#endif //__KLFIXEDBASE_ARM_HEADERFILE__
