/**************************************************************************\
 *
 *  This file is part of the Klimt library.
 *  Copyright (C) 2003 by IMS, Vienna University of Technology.
 *  All rights reserved.
 *
 *  This product is based on the published OpenGL ES API, but is not an
 *  implementation which is certified or licensed by Silicon Graphics, Inc.
 *  under the OpenGL ES API.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  ("GPL") version 2 as published by the Free Software Foundation.
 *  See the file LICENSE.GPL at the root directory of this source
 *  distribution for additional information about the GNU GPL.
 *  For the full GPL license see
 *  <URL:http://www.gnu.org/copyleft/gpl.html>
 *
 *  For using Klimt with software that can not be combined with the
 *  GNU GPL, and for taking advantage of the additional benefits of
 *  our support services, please contact IMS about acquiring a
 *  Klimt Professional Edition License.
 *
 *  Contact: <mailto:klimt@studierstube.org>
 *  See <URL:http://www.studierstube.org/klimt>
 *  for more information.
 *
 *  Vienna University of Technology
 *  Institute for Software Technology and Interactive Systems
 *  Interactive Media Systems Group
 *  Favoritenstrasse 9-11/188/2
 *  A-1040 Vienna, Austria
 *  <URL:http://www.ims.tuwien.ac.at>.
 *
 **************************************************************************
 *
 * $Header: /cvsroot/klimt/klimt/klimt/src/klGlu.cpp,v 1.8 2004/06/15 10:15:49 wasserer Exp $
 *
\**************************************************************************/

#include "klMain.h"
#include "gl/gl.h"

#include <math.h>

struct GLUquadricObj;


// always use c name mangling convention
//
extern "C"
{


KLIMT_API void
gluPerspective(double fovy, double aspect, double zNear, double zFar)
{
   double xmin, xmax, ymin, ymax;

   ymax = zNear * tan(fovy * M_PI / 360.0f);
   ymin = -ymax;
   xmin = ymin * aspect;
   xmax = ymax * aspect;

   glFrustum(xmin, xmax, ymin, ymax, zNear, zFar);
}  


KLIMT_API const GLubyte*
gluErrorString(GLenum /*errCode*/)
{
	return (const GLubyte*)("Error Report not supported!");
}


KLIMT_API void
gluCylinder(GLUquadricObj * /*qobj*/, GLdouble baseRadius, GLdouble topRadius, GLdouble height, 
			GLint slices, GLint stacks)
{
	float t = 0.0f, z = 0.0f, r = (float)baseRadius, s, ds = 1.0f/slices, dt = 1.0f / stacks,
		  da = 2.0f * M_PI / slices,
		  dr = (float)((topRadius - baseRadius) / stacks),
		  dz = (float)height / stacks,
		  nz = (float)(baseRadius - topRadius) / (float)height;
	int   slice;

	while(stacks--)
	{
		glBegin(GL_QUAD_STRIP);
		for(slice=0, s=0.0f; slice<=slices; slice++, s+=ds)
		{
			float	phi = (slice==slices) ? 0.0f : slice*da,
					x = (float)sin(phi), y = (float)cos(phi), nzn = nz,
					len = (float)sqrt(x*x + y*y + nz*nz);

			if(len>0.0001f)
			{
				float invLen = 1.0f/len;
				x *= invLen;
				y *= invLen;
				nzn = nz*invLen;
			}

			glNormal3f(x,y,nzn);

			glTexCoord2f(s,t);
			glVertex3f(x*r, y*r, z);

			glTexCoord2f(s,t+dt);
			glVertex3f(x*(r+dr), y*(r+dr), z+dz);
		}
		glEnd();
		r += dr;	t += dt;	z += dz;
	}
}


KLIMT_API void
gluDisk(GLUquadricObj * /*qobj*/, GLdouble innerRadius, GLdouble outerRadius, GLint slices, GLint loops)
{
	float da = 2.0f * M_PI / slices,
		  dr = (float)(outerRadius - innerRadius) / loops,
		  r1 = (float)innerRadius,
		  invOut = 2.0f/(float)outerRadius;

	glNormal3f(0.0f, 0.0f, 1.0f);

	while(loops--)
	{
		float r2 = r1 + dr,
			  sr1 = r1*invOut, sr2 = r2*invOut;

		glBegin(GL_QUAD_STRIP);
		for(int s=0; s<=slices; s++)
		{
			float a = (s==slices) ? 0.0f : s*da,
				  sinv = (float)sin(a), cosv = (float)cos(a);

			glTexCoord2f(0.5f+sr2*sinv, 0.5f+sr2*cosv);
			glVertex2f(r2*sinv, r2*cosv);

			glTexCoord2f(0.5f+sr1*sinv, 0.5f+sr1*cosv);
			glVertex2f(r1*sinv, r1*cosv);
		}
		glEnd();

		r1 = r2;
	}
}


#define setFloat3(f, x,y,z) f[0]=(x);  f[1]=(y);  f[2]=(z);
#define cpyFloat3(to, from) to[0]=from[0];  to[1]=from[1];  to[2]=from[2];
#define mulFloat3(f, v) f[0]*=(v);  f[1]*=(v);  f[2]*=(v);

KLIMT_API void
gluSphere(GLUquadricObj * /*qobj*/, GLdouble radius, GLint slices, GLint stacks)
{
	if(stacks<3) stacks = 3;
	if(slices<4) slices = 4;
	if(slices>64) slices = 64;

	float coords[129][3], normals[129][3], S[129], tmp[3];
	float rho, drho, theta, dtheta, tc, ts;
	int i, j;

	drho = float(M_PI) / (float) (stacks-1);
	dtheta = 2.0f * float(M_PI) / (float) slices;

	float currs = 0.0f;
	float incs = 1.0f / (float)slices;
	rho = drho;
	theta = 0.0f;
	tc = (float)cos(rho);
	ts = -(float)sin(rho);

	setFloat3(tmp, 0.0f, tc, ts);
	cpyFloat3(normals[0], tmp);
	mulFloat3(tmp, (float)radius);
	cpyFloat3(coords[0], tmp);
	S[0] = currs;

	float dT = 1.0f / (float) (stacks-1);
	float T = 1.0f - dT;

	glBegin(GL_TRIANGLES);
	for (j = 1; j <= slices; j++) {
		glNormal3f(0.0f, 1.0f, 0.0f);
		glTexCoord2f(currs + 0.5f * incs, 1.0f);
		glVertex3f(0.0f, (float)radius, 0.0f);

		glNormal3fv(normals[j-1]);
		glTexCoord2f(currs, T);
		glVertex3fv(coords[j-1]);

		currs += incs;
		theta += dtheta;
		S[j] = currs;
		setFloat3(tmp, float(sin(theta))*ts, tc, float(cos(theta))*ts);

		cpyFloat3(normals[j], tmp);
		mulFloat3(tmp, (float)radius);
		cpyFloat3(coords[j], tmp);

		glNormal3fv(normals[j]);
		glTexCoord2f(currs, T);
		glVertex3fv(coords[j]);
	}
	glEnd();

	rho += drho;

	for(i = 2; i < stacks-1; i++) {
		tc = (float)cos(rho);
		ts = -(float) sin(rho);
		glBegin(GL_QUAD_STRIP);
		theta = 0.0f;
		for (j = 0; j <= slices; j++) {
			glNormal3fv(normals[j]);
			glTexCoord2f(S[j], T);
			glVertex3fv(coords[j]);

			setFloat3(tmp, float(sin(theta))*ts, tc, float(cos(theta))*ts);
			cpyFloat3(normals[j], tmp);
			glNormal3fv(tmp);
			glTexCoord2f(S[j], T-dT);
			mulFloat3(tmp, (float)radius);
			cpyFloat3(coords[j], tmp);
			theta += dtheta;
			glVertex3fv(tmp);
		}
		glEnd();
		rho += drho;
		T -= dT;
	}


	glBegin(GL_TRIANGLES);
	for (j = 0; j < slices; j++) {
		glTexCoord2f(S[j], T);
		glNormal3fv(normals[j]);
		glVertex3fv(coords[j]);

		glTexCoord2f(S[j]+incs*0.5f, 0.0f);
		glNormal3f(0.0f, -1.0f, 0.0f);
		glVertex3f(0.0f, -(float)radius, 0.0f);

		glTexCoord2f(S[j+1], T);
		glNormal3fv(normals[j+1]);
		glVertex3fv(coords[j+1]);
	}
	glEnd();
}


}	//extern "C"
